/*****************************************************************************
 *                                  _   _ ____  _     
 *  Project                     ___| | | |  _ \| |    
 *                             / __| | | | |_) | |    
 *                            | (__| |_| |  _ <| |___ 
 *                             \___|\___/|_| \_\_____|
 *
 * Copyright (C) 2001, Daniel Stenberg, <daniel@haxx.se>, et al.
 *
 * In order to be useful for every potential user, curl and libcurl are
 * dual-licensed under the MPL and the MIT/X-derivate licenses.
 *
 * You may opt to use, copy, modify, merge, publish, distribute and/or sell
 * copies of the Software, and permit persons to whom the Software is
 * furnished to do so, under the terms of the MPL or the MIT/X-derivate
 * licenses. You may pick one of these licenses.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 *****************************************************************************/

#include "setup.h"

#include <stdlib.h>

#include "export_to_user.h"
#include "sync.h"


#ifdef HAVE_PTHREAD_H
#  include <pthread.h>
#elif defined HAVE_WIN32_THREADS
#  include <windows.h>
#else
#  define NOT_SUPPORTED_THREAD_IMPL 1
#endif

/* The last #include file should be: */
#ifdef MALLOCDEBUG
#  include "memdebug.h"
#endif

/*
 * The prototypes for the implementations of the mutex functions:
 */
static curl_mutex_t* mutex_new_impl(void);
static void mutex_free_impl(curl_mutex_t*);
static void mutex_lock_impl(curl_mutex_t*);
static void mutex_unlock_impl(curl_mutex_t*);

/*
 * The mutex type that holds thread implementation specific data.
 */
struct _curl_mutex {
#ifdef HAVE_PTHREAD_H
  pthread_mutex_t lock;
#elif defined HAVE_WIN32_THREADS
  HANDLE lock;
#else
  int empty;
#endif
};


/*
 * Try to register the global mutex functions with libcURL.
 */
int curl_try_register_mutex_functions(void)
{
#ifdef NOT_SUPPORTED_THREAD_IMPL
  /* the thread implementation is not (yet) supported by libcURL */
  return (-1);
#else
  curl_mutex_functions_t* functions;
  
  /* create the struct with the function pointers an register it to libcURL */
  if ((functions = malloc(sizeof(curl_mutex_functions_t))) == NULL)
    return (1); /* out of memory */

  /* fill the function struct */
  functions->new_function= mutex_new_impl;
  functions->free_function= mutex_free_impl;
  functions->lock_function= mutex_lock_impl;
  functions->unlock_function= mutex_unlock_impl;

  /* ok, register with libcURL */
  curl_register_mutex_functions(functions);

  return (0); /* everything looks fine */
#endif  
}

/*
 * Unregister and delete the previously set mutex function struct
 */
void curl_try_unregister_mutex_functions(void)
{
  curl_mutex_functions_t* functions;
  
  /* get the registered mutex function struct */
  functions= curl_get_mutex_functions();
  
  /* unregister it */
  curl_unregister_mutex_functions();
  
  /* delete the struct */
  if (functions != NULL)
    free(functions);
}


/*
 * The actual mutex implementation
 */
curl_mutex_t* mutex_new_impl(void)
{
  curl_mutex_t* mutex;

  if ((mutex = malloc(sizeof(curl_mutex_t))) == NULL)
    return (NULL);  /* a error should be reported! Somehow... */
#ifdef HAVE_PTHREAD_H
  pthread_mutex_init(&mutex->lock, NULL);
#elif defined HAVE_WIN32_THREADS
  InitializeCriticalSection(mutex->lock);
#endif
  return (mutex);
}

void mutex_free_impl(curl_mutex_t* m)
{
#ifdef HAVE_PTHREAD_H
  pthread_mutex_destroy(&m->lock);
#elif defined HAVE_WIN32_THREADS
  DeleteCriticalSection(m->lock);
#endif
  free(m);
}

void mutex_lock_impl(curl_mutex_t* m)
{
#ifdef HAVE_PTHREAD_H
  pthread_mutex_lock(&m->lock);
#elif defined HAVE_WIN32_THREADS
  EnterCriticalSection(m->lock);
#endif
}

void mutex_unlock_impl(curl_mutex_t* m)
{
#ifdef HAVE_PTHREAD_H
  pthread_mutex_unlock(&m->lock);
#elif defined HAVE_WIN32_THREADS
  LeaveCriticalSection(m->lock);
#endif
}
